%% @hidden
-module(couchdb_ejson).

-export([encode/1, decode/1]).
-export([post_decode/1]).

-include("couchdb.hrl").
-include("./../../dev.hrl").

-ifndef('WITH_JIFFY').
% -define(JSON_ENCODE(D), jsx:encode(pre_encode(D))).
% -define(JSON_DECODE(D), post_decode(jsx:decode(D))).

-define(JSON_ENCODE(D), jsone:encode(pre_encode(D))).
-define(JSON_DECODE(D), post_decode(jsone:decode(D))).



-else.
-define(JSON_ENCODE(D), jiffy:encode(D, [uescape])).
-define(JSON_DECODE(D), jiffy:decode(D)).
-endif.


-spec encode(ejson()) -> binary().

%% @doc encode an erlang term to JSON. Throw an exception if there is
%% any error.
encode(D) ->
    ?JSON_ENCODE(D).

-spec decode(binary()) -> ejson().
%% @doc decode a binary to an EJSON term. Throw an exception if there is
%% any error.
decode(D) ->
    try
        ?JSON_DECODE(D)
    catch
        throw:Error ->
            throw({invalid_json, Error});
        error:badarg ->
            throw({invalid_json, badarg})
    end.


pre_encode(X) when is_map(X) ->    
    maps:map(fun(_Key, Value) ->        
        pre_encode(Value)    
    end, X);
pre_encode({[]}) ->
    [{}];
pre_encode({PropList}) ->
    pre_encode(PropList);
pre_encode([{_, _}|_] = PropList) ->
    [ {Key, pre_encode(Value)} || {Key, Value} <- PropList ];
pre_encode(List) when is_list(List) ->
    [ pre_encode(Term) || Term <- List ];
pre_encode(true) ->
    true;
pre_encode(false) ->
    false;
pre_encode(null) ->
    null;
pre_encode(nil) ->
    null;
pre_encode(Atom) when is_atom(Atom) ->
    erlang:atom_to_binary(Atom, utf8);
pre_encode(<<Any/binary>>) ->
    unicode:characters_to_binary(Any, latin1, utf8);
pre_encode(Term) when is_integer(Term); is_float(Term); is_binary(Term) ->
    Term;

%  Irsan added: basic map
pre_encode(#{} = Data) -> Data;
% Irsan added: List of maps
pre_encode([#{}, _] = Data) -> Data. 


post_decode(X) when is_map(X) ->    
    maps:map(fun(_Key, Value) ->
        post_decode(Value)    
    end, X);
post_decode({[{}]}) ->
    {[]};
post_decode([{}]) ->
    {[]};
post_decode([{_Key, _Value} | _Rest] = PropList) ->
    {[ {Key, post_decode(Value)} || {Key, Value} <- PropList ]};
post_decode(List) when is_list(List) ->
    [ post_decode(Term) || Term <- List];
post_decode({Term}) ->
    post_decode(Term);
post_decode(null) ->
    nil;
post_decode(Term) ->
    Term.
